/*******************************************************************************
* File Name    : LCD_LIB
* Version      : Rev. 2.00
* Device(s)    : R7F100GL
* Tool-Chain   : Arduino
* Description  : This file implements LCD control function for ACM1602
* Creation Date: 2016/09/30 First Release
* Modified Date: 2022/05/20 Ported to RL78/G23-64PFPB
*                2023/04/22 Ported to RL78/G22-48PFPB
********************************************************************************/
#include "LCD_LIB.h"

#define SEL_COMMAND         (0U)        // parameter for RW pin to set command 
#define SEL_DATA            (1U)        // parameter for RW pin to set data

/* Maximum characters per line of LCD display. */
#define LCD_SIZE            16
/* number of lines on the LCD display */
#define MAXIMUM_LINES       2

/* LCD data address for display (MAX 40 charactor/line)  */
#define LCD_LINE1 0x00                  /* first line start from address 0x00   */
#define LCD_LINE2 0x40                  /* 2nd line start from address 0x40     */


/*****************************************************************************
LCD commands - use LCD_write function to write these commands to the LCD. 
*****************************************************************************/
/* Clear LCD display (set 20H) and home cursor              wait 1.64ms */
#define LCD_CLEAR        0b00000001
/*                                +--:clear display command */

/* Cursor auto decrement after R/W (not shift display)      wait 40us   */
#define CURSOR_MODE_DEC  0b00000100

/* Cursor auto increment after R/W (not shift display)  */
#define CURSOR_MODE_INC  0b00000110
/*                              ||+--:SH:shift display position
                                ||  * 0:not shift
                                ||  * 1:shift display
                                |+---:I/D:Incement/Decrement pointer
                                |     0:move pointer to left 
                                |   * 1:move pointer to right
                                +----:entry mode set command    */

/* Display ON with Cursor                                   wait 40us   */
#define LCD_CURSOR_ON    0b00001110
/* Display ON with Cursor off   */
#define LCD_CURSOR_OFF   0b00001100
/* Display on with blinking cursor  */
#define LCD_CURSOR_BLINK 0b00001101
/*                             |||+--:B:Cursor blink
                               |||  * 0:blink off
                               |||  * 1:blink on
                               ||+---:C:Cursor ON/OFF
                               ||   * 0:disappered
                               ||   * 1:cursor is turned on
                               |+----:D:Display ON/OFF
                               |      0:display is turned off
                               |    * 1:display is turned on
                               +-----Display ON/OFF command */

/*Move Cursor Left One Position                             wait 40us   */
#define LCD_CURSOR_LEFT  0b00010000
/* Move Cursor Right One Position */
#define LCD_CURSOR_RIGHT 0b00010100
/* Move Display Right One Position */
#define LCD_DISP_LEFT    0b00011000
/* Move Display Right One Position */
#define LCD_DISP_RIGHT   0b00011100
/*                            |++----:S/C:R/L:Cursor or display shift mode
                              |    * 00:shift cursor to the left
                              |    * 01:shift cursor to the right
                              |      10:shift display and cursor to the left
                              |      11:shift display and cursor to the right
                              +-----Cursor or display shift command */

/* Setup, 4 bits bus,2 line display, 5x7 font               wait 40us   */
#define FUNCTION_SET     0b00101000
/*                           |||+----:F:Font type
                             |||    * 0:5*8 dot font
                             |||      1:5*11 dot font
                             ||+-----:N:Display line number
                             ||       0:1-line mode
                             ||     * 1:2-line mode
                             |+------:DL:Interface data length
                             |      * 0:4bit(D7-D4)bus mode
                             |        1:8bit(D7-D0)bus mode
                             +-------:Function set command  */

/* 1st nibble for 4bit I/F mode     */
#define F_4bitIF         0x02U
/* 2nd nibble for 4bit I/F mode     */
#define LCD_TWO_LINE     0x08U

/* move cursor to top of line 1 (data pointer)  */
#define LCD_HOME_L1      0b10000000
/*                         |+++++++--:data pointer(0x00:top of line1)
                           +---------:pointer set command   */
/* move cursor to top of line 2 (data pointer)  */
#define LCD_HOME_L2      0b11000000
/*                         |+++++++--:data pointer(0x40:top of line2)
                           +---------:pointer set command   */


/*----------------------------------------------
    LCD pin definition
------------------------------------------------*/
volatile uint8_t rs_pin;                    // RS信号の端子を指定する変数
volatile uint8_t rw_pin;                    // RW信号の端子を指定する変数
volatile uint8_t e_pin;                     // Enable信号の端子を指定する変数
volatile uint8_t d4_pin;                    // D4信号の端子を指定する変数
volatile uint8_t d5_pin;                    // D5信号の端子を指定する変数
volatile uint8_t d6_pin;                    // D6信号の端子を指定する変数
volatile uint8_t d7_pin;                    // D7信号の端子を指定する変数

/*******************************************************************************
* Function Name: init_LCD
* Description  : この関数で、LCDCを4ビットI/F、2行表示に設定
* Arguments    : 使用する端子
*               1st argument : RS信号の端子の指定
*               2nd argument : RW信号の端子の指定
*               3rd argument : Enable信号の端子の指定
*               4th argument : データ4(D4)の端子の指定
*               5th argument : データ5(D5)の端子の指定
*               6th argument : データ6(D6)の端子の指定
*               7th argument : データ7(D7)の端子の指定
* Return Value : none
********************************************************************************/
void init_LCD(uint8_t pin_rs, uint8_t pin_rw, uint8_t pin_e,
               uint8_t pin_d4, uint8_t pin_d5,
               uint8_t pin_d6, uint8_t pin_d7)
{

/*----------------------------------------------
    set pin assignment for LCD
------------------------------------------------*/
    rs_pin = pin_rs;                    // RS信号の端子を設定
    rw_pin = pin_rw;                    // RW信号の端子を設定
    e_pin  = pin_e;                     // Enable信号の端子を設定
    d4_pin = pin_d4;                    // D4信号の端子を設定
    d5_pin = pin_d5;                    // D5信号の端子を設定
    d6_pin = pin_d6;                    // D6信号の端子を設定
    d7_pin = pin_d7;                    // D7信号の端子を設定

/*----------------------------------------------
    assign pin as output mode
------------------------------------------------*/

    pinMode(rs_pin, OUTPUT);            // RS信号端子を出力に設定
    pinMode(rw_pin, OUTPUT);            // RW信号端子を出力に設定
    pinMode(e_pin, OUTPUT);             // Enable信号端子を出力に設定
    digitalWrite(e_pin, LOW);           // Enable信号をネゲートする
    pinMode(d4_pin, OUTPUT);            // D4信号端子を出力に設定
    pinMode(d5_pin, OUTPUT);            // D5信号端子を出力に設定
    pinMode(d6_pin, OUTPUT);            // D6信号端子を出力に設定
    pinMode(d7_pin, OUTPUT);            // D7信号端子を出力に設定


/*==========================================
    initialise LCDC(HD44780 or equivalent)
============================================*/

    delay(100);                         // LCDC電源立ち上げから100ms待つ
    digitalWrite(rs_pin, SEL_COMMAND);  // 制御信号選択側に設定
    digitalWrite(rw_pin, LOW);          // リード／ライトをライトに設定

/*----------------------------------------------
    set 4bit interface mode(preprocessing)
    LCDが4/8bitI/Fモードでも一旦8bitにしてから4bitにする
------------------------------------------------*/

    LCD_nibble_write(0x03U);            // 8bitI/Fモードに設定
    delay(5);                           // 5ミリ秒待つ
    LCD_nibble_write(0x03U);            // 4bitモード時のダミー
    delayMicroseconds(100);             // 100マイクロ秒待つ
    LCD_nibble_write(0x03U);            // 4bitモード時のダミー
    delayMicroseconds(50);              // 50マイクロ秒待つ

/*----------------------------------------------
    set  high nibble of 4bit I/F command only
    I/Fだけを4ビットに変更（下位4ビットはまだ不定）
------------------------------------------------*/

    LCD_nibble_write(F_4bitIF);         // 4bitI/Fモードに設定
    delayMicroseconds(50);              // 50マイクロ秒待つ

/*----------------------------------------------
    set complete 4bit I/F command
    ここからは，4ビットI/FでI/F条件を正しく設定する
------------------------------------------------*/

    /* LCDの構成（I/Fのビット長，行数，ドット数）の指定 */
    LCD_nibble_write(F_4bitIF);         // LCDの指定で，再度設定
    LCD_nibble_write(LCD_TWO_LINE);     // 2行表示
    delayMicroseconds(50);              // 50マイクロ秒待つ

    /* 表示状態の設定                   */
    LCD_write(LCD_CURSOR_OFF);          // カーソル非表示でLCD表示
    delayMicroseconds(50);              // 50マイクロ秒待つ

    /* LCD表示をクリアする              */
    LCD_write(LCD_CLEAR);               // LCDのクリアコマンド送信
    delayMicroseconds(2100U);           // クリア処理時間待ち(2.1ms)

    /* データ入力モード設定             */
    LCD_write(CURSOR_MODE_INC);         // 書き込みでカーソルを右に移動
    delayMicroseconds(50);              // 50マイクロ秒待つ
    LCD_write(LCD_HOME_L1);             // カーソルを1行目の1文字目に移動
    delayMicroseconds(50);              // 50マイクロ秒待つ

}
/******************************************************************************
End of function init_LCD function
******************************************************************************/


/*******************************************************************************
* Function Name: setCursor
* Description  : この関数でカーソル位置を指定する
* Arguments    : cursor address
*               col: 0 to 0x0F
*               row : 0(line1) /1(line2)
* Return Value : none
********************************************************************************/
void setCursor( uint8_t col , uint8_t row)
{

    uint8_t work;

    work = (( row & 1 ) * 0x40);            // 行の先頭アドレス作成
    work += ( col & 0x0F );                 // カーソル位置を作成
    work |= 0x80;                           // カーソル移動コマンド作成

    LCD_write( work );                      // カーソルを移動
    delayMicroseconds(60);                  // 60マイクロ秒待つ

}
/******************************************************************************
End of function setCursor
******************************************************************************/


/*******************************************************************************
* Function Name: set_2digitBCD
* Description  : この関数は引数（99-0）を十進数としてASCII変換して、2桁で表示
* Arguments    : display number(99-0)
* Return Value : none
********************************************************************************/
void set_2digitBCD( uint8_t data)
{

    uint8_t work;

    work = data;                            // 表示データをセット
    
    set_1digit((work / 10));                // 10の桁を表示
    delayMicroseconds(60);                  // 60マイクロ秒待つ
    set_1digit((work % 10));                // 1の桁を表示
    delayMicroseconds(60);                  // 60マイクロ秒待つ

}
/******************************************************************************
End of function set_2digitBCD
******************************************************************************/


/*******************************************************************************
* Function Name: set_2digit
* Description  : この関数は引数（8ビット）をASCII変換して2桁で表示
* Arguments    : display number(packed BCD or HEX)
* Return Value : none
********************************************************************************/
void set_2digit( uint8_t data)
{

    uint8_t work;

    work = data;                            // 表示データをセット

    set_1digit( work >> 4 );                // 上位4bitを表示
    delayMicroseconds(60);                  // 60マイクロ秒待つ
    set_1digit( data );                     // 下位4bitを表示
    delayMicroseconds(60);                  // 60マイクロ秒待つ

}
/******************************************************************************
End of function set_2digit
******************************************************************************/


/*******************************************************************************
* Function Name: set_1digit
* Description  : この関数は引数の下位4ビットをASCII変換して表示
* Arguments    : display lowwer 4bit as ASCII
* Return Value : none
********************************************************************************/
void set_1digit( uint8_t data)
{

    uint8_t work;

    work = (data & 0x0F);                   // 未使用ビットをマスク
    if ( work < 0x0A )
    {
        work += 0x30;                       // 0-9をASCIIコードに変換
    }
    else
    {
        work += 0x37;                       // A-FをASCIIコードに変換
    }

    LCD_DATA_write(work);                   // 1桁のデータ（0-F）を表示

}
/******************************************************************************
End of function set_1digit
******************************************************************************/


/*******************************************************************************
* Function name     : DisplayString
* Description       : この関数はLCDの第1引数で示す位置から第2引数で示すデータを表示する
*                     第1引数は0-63なら1行目、64-127なら2行目を示す
*                     表示データはASCIIコードとし、0x00で終端する
*                     For example, to start at the 2nd position on line 1...
*                     DisplayString(LCD_LINE1 + 1, string address) 
* Argument          : (uint8_t) LCD position 
*                     (int8_t *) string
* Return value      : none.
* Remarks           : 表示用RAM領域は2行なら 1行目:0-27H(=39）,2行目:40H(=64)-67H(103)
*
*                     4行表示LCDでは、1行目:0-13H(19)、2行目:40H(=64)-53H(83)、
*                                     3行目:14H(20)-27H(39)、4行目:54H(84)-67H(103)
*******************************************************************************/
void DisplayString(uint8_t position, int8_t * string)
{

    if ( position < LCD_LINE2 )
    {                                       // カーソルをline1に移動させる
        /* Display on Line 1 */
        LCD_write((uint8_t)(LCD_HOME_L1 + position));
        delayMicroseconds(60);              // 60マイクロ秒待つ
    }
    else
    {                                       // カーソルをline2に移動させる
        /* Display on Line 2 */
        LCD_write((uint8_t)(LCD_HOME_L2 + (position - LCD_LINE2)));
        delayMicroseconds(60);              // 60マイクロ秒待つ
    }

    digitalWrite(rs_pin, SEL_DATA);         // データを選択

    do
    {
        LCD_write(*string++);               // 表示データ転送
        delayMicroseconds(60);              // 60マイクロ秒待つ

    }
    while (*string);

    digitalWrite(rs_pin, SEL_COMMAND);      // コマンド転送に戻す

}
/******************************************************************************
End of function DisplayString
******************************************************************************/


/******************************************************************************
* Function name     : LCD_DATA_write
* Description       : This function write data to  LCD
* Argument          : (uint8_t)value - the data to write
* Return value      : none
*******************************************************************************/
void LCD_DATA_write(uint8_t value)
{

    digitalWrite(rs_pin, SEL_DATA);         // データを選択

    /* Write upper nibble first */
    LCD_nibble_write((uint8_t)(value >> 4U));

    /* Write lower nibble second */
    LCD_nibble_write( value );

    digitalWrite(rs_pin, SEL_COMMAND);      // コマンド転送に戻す

}

/******************************************************************************
End of function LCD_DATA_write
******************************************************************************/

/******************************************************************************
* Function name     : LCD_write
* Description       : この関数は引数をコマンドとして書き込むD
* Argument          : (uint8_t)value - コマンド
* Return value      : none
*******************************************************************************/
void LCD_write(uint8_t value)
{
    /* Write upper nibble first */
    LCD_nibble_write((uint8_t)(value >> 4));    // 上位4ビットをLCDに書き込む

    /* Write lower nibble second */
    LCD_nibble_write( value );                  // 下位4ビットルをLCDに書き込む

}
/******************************************************************************
End of function LCD_write
******************************************************************************/

/*******************************************************************************
* Function name : LCD_write_nibble
* Description   : この関数は引数の下位4ビットをLCDに書き込む
* Argument      : (uint8_t) value - the value to write(low nibble)
* Return value  : none
*******************************************************************************/
void LCD_nibble_write(uint8_t value)
{
    uint8_t work = 0U;

    /* EN enable chip (HIGH) */
    digitalWrite(e_pin, HIGH);           // 書き込み信号をアサート

    /* Tiny delay */

    NOP();

    work = value;                       // 設定値の読み出し
    digitalWrite(d4_pin, ( work & 1 )); // D4(LSB)を設定
    work >>= 1;                         // 次のビットの準備
    digitalWrite(d5_pin, ( work & 1 )); // D5を設定
    work >>= 1;                         // 次のビットの準備
    digitalWrite(d6_pin, ( work & 1 )); // D6を設定
    work >>= 1;                         // 次のビットの準備
    digitalWrite(d7_pin, ( work & 1 )); // D7を設定
    

    /* データのセットアップ時間を確保   */
    delayMicroseconds(2);               // マージンを含めて2us確保

    digitalWrite(e_pin, LOW);           // 書き込み信号をネゲート

    delayMicroseconds(2);               // 2マイクロ秒待つ

}
/*******************************************************************************
End of function LCD_nibble_write
*******************************************************************************/
